from __future__ import annotations

from datetime import datetime
from typing import Any, Literal
from uuid import UUID

from pydantic import Field

from pyhuntress.models.base.huntress_model import HuntressModel

class SIEMPagination(HuntressModel):
    current_page: int | None = Field(default=None, alias="CurrentPage")
    current_page_count: int | None = Field(default=None, alias="CurrentPageCount")
    limit: int | None = Field(default=None, alias="Limit")
    total_count: int | None = Field(default=None, alias="TotalCount")
    next_page: int | None = Field(default=None, alias="NextPage")
    next_page_url: str | None = Field(default=None, alias="NextPageURL")
    next_page_token: str | None = Field(default=None, alias="NextPageToken")
    
class SIEMAgents(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    version: str | None = Field(default=None, alias="Version")
    arch: str | None = Field(default=None, alias="Arch")
    win_build_number: int | None = Field(default=None, alias="WinBuildNumber")
    domain_name: str | None = Field(default=None, alias="DomainName")
    created_at: datetime | None = Field(default=None, alias="CreateAt")
    hostname: str | None = Field(default=None, alias="Hostname")
    ipv4_address: str | None = Field(default=None, alias="IPv4Address")
    external_ip: str | None = Field(default=None, alias="ExternalIP")
    mac_addresses: list | None = Field(default=None, alias="MacAddresses")
    updated_at: datetime | None = Field(default=None, alias="IPv4Address")
    last_survey_at: datetime | None = Field(default=None, alias="LastSurveyAt")
    last_callback_at: datetime | None = Field(default=None, alias="LastCallbackAt")
    account_id: int | None = Field(default=None, alias="AccountID")
    organization_id: int | None = Field(default=None, alias="OrganizationID")
    platform: Literal[
        "windows",
        "darwin",
        "linux",
    ] | None = Field(default=None, alias="Platform")
    os: str | None = Field(default=None, alias="OS")
    service_pack_major: int | None = Field(default=None, alias="ServicePackMajor")
    service_pack_minor: int | None = Field(default=None, alias="ServicePackMinor")
    tags: list | None = Field(default=None, alias="Tags")
    os_major: int | None = Field(default=None, alias="OSMajor")
    os_minor: int | None = Field(default=None, alias="OSMinor")
    os_patch: int | None = Field(default=None, alias="OSPatch")
    version_number: int | None = Field(default=None, alias="VersionNumber")
    edr_version: str | None = Field(default=None, alias="EDRVersion")
    os_build_version: str | None = Field(default=None, alias="OSBuildVersion")
    serial_number: str | None = Field(default=None, alias="SerialNumber")
    defender_status: str | None = Field(default=None, alias="DefenderStatus")
    defender_substatus: str | None = Field(default=None, alias="DefenderSubstatus")
    defender_policy_status: str | None = Field(default=None, alias="DefenderPolicyStatus")
    firewall_status: str | None = Field(default=None, alias="FirewallStatus")

class SIEMAgentsResponse(HuntressModel):
    agents: dict[str, Any] | None = Field(default=None, alias="Agents")
    pagination: dict[str, Any] | None = Field(default=None, alias="Pagination")
    
class SIEMAccount(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    name: str | None = Field(default=None, alias="Name")
    subdomain: str | None = Field(default=None, alias="Subdomain")
    status: Literal[
        "enabled",
        "disabled",
    ] | None = Field(default=None, alias="Status")

class SIEMActorResponse(HuntressModel):
    account: dict[str, Any] | None = Field(default=None, alias="Account")
    user: str | None = Field(default=None, alias="User")

class SIEMBillingReports(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    plan: str | None = Field(default=None, alias="Plan")
    quantity: int | None = Field(default=None, alias="Quantity")
    amount: int | None = Field(default=None, alias="Amount")
    currency_type: str | None = Field(default=None, alias="CurrencyType")
    receipt: str | None = Field(default=None, alias="Receipt")
    status: Literal[
        "open",
        "paid",
        "failed",
        "partial_refund",
        "full_refund",
        "draft",
        "voided",
    ] | None = Field(default=None, alias="Status")
    created_at: datetime | None = Field(default=None, alias="CreatedAt")
    updated_at: datetime | None = Field(default=None, alias="UpdatedAt")

class SIEMBillingReportsResponse(HuntressModel):
    billing_reports: dict[str, Any] | None = Field(default=None, alias="BillingReports")

class SIEMIncidentReportsResponse(HuntressModel):
    incident_reports: dict[str, Any] | None = Field(default=None, alias="IncidentReports")
    pagination: dict[str, Any] | None = Field(default=None, alias="Pagination")
    
class SIEMIncidentReports(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    status: Literal[
        "sent",
        "closed",
        "dismissed",
        "auto_remediating",
        "deleting",
    ] | None = Field(default=None, alias="Status")
    summary: str | None = Field(default=None, alias="Summary")
    body: str | None = Field(default=None, alias="Body")
    updated_at: datetime | None = Field(default=None, alias="UpdatedAt")
    agent_id: int | None = Field(default=None, alias="AgentId")
    platform: Literal[
        "windows",
        "darwin",
        "microsoft_365",
        "google",
        "linux",
        "other",
    ] | None = Field(default=None, alias="Platform")
    status_updated_at: datetime | None = Field(default=None, alias="StatusUpdatedAt")
    organization_id: int | None = Field(default=None, alias="OrganizationId")
    sent_at: datetime | None = Field(default=None, alias="SentAt")
    account_id: int | None = Field(default=None, alias="AccountId")
    subject: str | None = Field(default=None, alias="Subject")
    remediations: list[dict[str, Any]] | None = Field(default=None, alias="Remediations")
    severity: Literal[
        "low",
        "high",
        "critical",
    ] | None = Field(default=None, alias="Severity")
    closed_at: datetime | None = Field(default=None, alias="ClosedAt")
    indicator_types: list | None = Field(default=None, alias="IndicatorTypes")
    indicator_counts: dict[str, Any] | None = Field(default=None, alias="IndicatorCounts")
    

class SIEMRemediations(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    type: str | None = Field(default=None, alias="Type")
    status: str | None = Field(default=None, alias="Status")
    details: dict[str, Any] | None = Field(default=None, alias="Details")
    completable_by_task_response: bool | None = Field(default=None, alias="CompletedByTaskResponse")
    completable_manually: bool | None = Field(default=None, alias="CompletedManually")
    display_action: str | None = Field(default=None, alias="DisplayAction")
    approved_at: datetime | None = Field(default=None, alias="ApprovedAt")
    approved_by: dict[str, Any] | None = Field(default=None, alias="ApprovedBy")
    completed_at: datetime | None = Field(default=None, alias="CompletedAt")
    
class SIEMRemediationsDetails(HuntressModel):
    rule_id: int | None = Field(default=None, alias="RuleId")
    rule_name: str | None = Field(default=None, alias="RuleName")
    completed_at: datetime | None = Field(default=None, alias="CompletedAt")
    forward_from: str | None = Field(default=None, alias="ForwardFrom")
    remediation: str | None = Field(default=None, alias="remediation")
    
class SIEMRemediationsApprovedBy(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    email: str | None = Field(default=None, alias="Email")
    first_name: str | None = Field(default=None, alias="FirstName")
    last_name: str | None = Field(default=None, alias="LastName")
    
class SIEMIndicatorCounts(HuntressModel):
    footholds: int | None = Field(default=None, alias="Footholds")
    mde_detections: int | None = Field(default=None, alias="MDEDetections")
    monitored_files: int | None = Field(default=None, alias="MonitoredFiles")
    siem_detections: int | None = Field(default=None, alias="SIEMDetections")
    managed_identity: int | None = Field(default=None, alias="ManagedIdentity")
    process_detections: int | None = Field(default=None, alias="ProcessDetections")
    ransomware_canaries: int | None = Field(default=None, alias="RansomwareCanaries")
    antivirus_detections: int | None = Field(default=None, alias="AntivirusDetections")
    
class SIEMOrganizationsResponse(HuntressModel):
    organizations: dict[str, Any] | None = Field(default=None, alias="Organizations")
    pagination: dict[str, Any] | None = Field(default=None, alias="Pagination")
    
class SIEMOrganizations(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    name: str | None = Field(default=None, alias="Name")
    created_at: datetime | None = Field(default=None, alias="CreatedAt")
    updated_at: datetime | None = Field(default=None, alias="UpdatedAt")
    account_id: int | None = Field(default=None, alias="AccountId")
    key: str | None = Field(default=None, alias="Key")
    notify_emails: list | None = Field(default=None, alias="NotifyEmails")
    microsoft_365_tenant_id: str | None = Field(default=None, alias="Microsoft365TenantId")
    incident_reports_count: int | None = Field(default=None, alias="IncidentsReportsCount")
    agents_count: int | None = Field(default=None, alias="AgentsCount")
    microsoft_365_users_count: int | None = Field(default=None, alias="Microsoft365UsersCount")
    sat_learner_count: int | None = Field(default=None, alias="SATLearnerCount")
    logs_sources_count: int | None = Field(default=None, alias="LogsSourcesCount")

class SIEMReportsResponse(HuntressModel):
    reports: dict[str, Any] | None = Field(default=None, alias="Organizations")
    pagination: dict[str, Any] | None = Field(default=None, alias="Pagination")

class SIEMReports(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    type: Literal[
        "monthly_summary",
        "quarterly_summary",
        "yearly_summary",
    ] | None = Field(default=None, alias="Type")
    period: str | None = Field(default=None, alias="Period")
    organization_id: int | None = Field(default=None, alias="OrganizationId")
    created_at: datetime | None = Field(default=None, alias="CreatedAt")
    updated_at: datetime | None = Field(default=None, alias="UpdatedAt")
    url: str | None = Field(default=None, alias="Type")
    events_analyzed: int | None = Field(default=None, alias="EventsAnalyzed")
    total_entities: int | None = Field(default=None, alias="TotalEntities")
    signals_detected: int | None = Field(default=None, alias="SignalsDetected")
    signals_investigated: int | None = Field(default=None, alias="SignalsInvestigated")
    itdr_entities: int | None = Field(default=None, alias="ITDREntities")
    itdr_events: int | None = Field(default=None, alias="ITDREvents")
    siem_total_logs: int | None = Field(default=None, alias="SIEMTotalLogs")
    siem_ingested_logs: int | None = Field(default=None, alias="SIEMIngestedLogs")
    autorun_events: int | None = Field(default=None, alias="AutorunEvents")
    autorun_signals_detected: int | None = Field(default=None, alias="AutorunSignalsDetected")
    investigations_completed: int | None = Field(default=None, alias="InvestigationsCompleted")
    autorun_signals_reviewed: int | None = Field(default=None, alias="AutorunSignalsReviewed")
    incidents_reported: int | None = Field(default=None, alias="IncidentsReported")
    itdr_incidents_reported: int | None = Field(default=None, alias="ITDRIncidentsReported")
    siem_incidents_reported: int | None = Field(default=None, alias="SIEMIncidentsReported")
    incidents_resolved: int | None = Field(default=None, alias="IncidentsResolved")
    incident_severity_counts: dict[str, int] | None = Field(default=None, alias="IncidentSeverityCounts")
    incident_product_counts: dict[str, int] | None = Field(default=None, alias="IncidentProductCounts")
    incident_indicator_counts: dict[str, int] | None = Field(default=None, alias="IncidentIndicatorCounts")
    top_incident_av_threats: list | None = Field(default=None, alias="TopIncidentAVThreats")
    top_incident_hosts: Any | None = Field(default=None, alias="TopIncidentHosts") #Huntress seems inconsistent between list and dict here
    potential_threat_indicators: int | None = Field(default=None, alias="PotentialThreatIndicators")
    agents_count: int | None = Field(default=None, alias="AgentsCount")
    deployed_canaries_count: int | None = Field(default=None, alias="DeployedCanariesCount")
    protected_profiles_count: int | None = Field(default=None, alias="ProtectedProfilesCount")
    windows_agent_count: int | None = Field(default=None, alias="WindowsAgentCount")
    macos_agent_count: int | None = Field(default=None, alias="MacOSAgentCount")
    servers_agent_count: int | None = Field(default=None, alias="ServersAgentCount")
    analyst_note: str | None = Field(default=None, alias="AnalystNote")
    global_threats_note: str | None = Field(default=None, alias="GlobalThreatsNote")
    ransomware_note: str | None = Field(default=None, alias="RansomwareNote")
    incident_log: list[dict[str, Any]] | None = Field(default=None, alias="IncidentLog")
    total_mav_detection_count: int | None = Field(default=None, alias="TotalMAVDetectionCount")
    blocked_malware_count: int | None = Field(default=None, alias="BlockedMalwareCount")
    investigated_mav_detection_count: int | None = Field(default=None, alias="InvestigatedMAVDetectionCount")
    mav_incident_report_count: int | None = Field(default=None, alias="MAVIncidentReportCount")
    autoruns_reviewed: int | None = Field(default=None, alias="AutorunsReviewed")
    host_processes_analyzed: int | None = Field(default=None, alias="HostProcessesAnalyzed")
    process_detections: int | None = Field(default=None, alias="ProcessDetections")
    process_detections_reviewed: int | None = Field(default=None, alias="ProcessDetectionsReviewed")
    process_detections_reported: int | None = Field(default=None, alias="ProcessDetectionsReported")
    itdr_signals: int | None = Field(default=None, alias="ITDRSignals")
    siem_signals: int | None = Field(default=None, alias="SIEMSignals")
    itdr_investigations_completed: int | None = Field(default=None, alias="ITDRInvestigationsCompleted")
    macos_agents: bool | None = Field(default=None, alias="MacOSAgents")
    windows_agents: bool | None = Field(default=None, alias="WindowsAgents")
    only_macos_agents: bool | None = Field(default=None, alias="OnlyMacOSAgents")
    antivirus_exclusions_count: int | None = Field(default=None, alias="AntivirusExclusionsCount")
    new_exclusions_count: int | None = Field(default=None, alias="NewExclusionsCount")
    allowed_exclusions_count: int | None = Field(default=None, alias="AllowedExclusionsCount")
    risky_exclusions_removed_count: int | None = Field(default=None, alias="RiskyExclusionsRemovedCount")

class SIEMSignalsResponse(HuntressModel):
    signals: dict[str, Any] | None = Field(default=None, alias="Organizations")
    pagination: dict[str, Any] | None = Field(default=None, alias="Pagination")

class SIEMSignals(HuntressModel):
    created_at: datetime | None = Field(default=None, alias="CreatedAt")
    id: int | None = Field(default=None, alias="Id")
    status: str | None = Field(default=None, alias="Status")
    updated_at: datetime | None = Field(default=None, alias="UpdatedAt")
    details: dict[str, Any] | None = Field(default=None, alias="Details")
    entity: dict[str, Any] | None = Field(default=None, alias="Entity")
    investigated_at: datetime | None = Field(default=None, alias="InvestigatedAt")
    investigation_context: str | None = Field(default=None, alias="InvestigationContext")
    name: str | None = Field(default=None, alias="Name")
    organization: dict[str, Any] | None = Field(default=None, alias="Organization")
    type: str | None = Field(default=None, alias="Type")
    
class SIEMSignalsDetails(HuntressModel):
    identity: str | None = Field(default=None, alias="Identity")
    application: str | None = Field(default=None, alias="Application")
    detected_at: datetime | None = Field(default=None, alias="DetectedAt")
    
class SIEMSignalsEntity(HuntressModel):
    id: int | None = Field(default=None, alias="Id")
    name: str | None = Field(default=None, alias="Name")
    type: Literal[
        "user_entity",
        "source",
        "mailbox",
        "service_principal",
        "agent",
        "identity",
    ] | None = Field(default=None, alias="Type")



import typing
from datetime import datetime, timezone
import base64

from pyhuntress.clients.huntress_client import HuntressClient
from pyhuntress.config import Config

if typing.TYPE_CHECKING:
    from pyhuntress.endpoints.siem.AccountEndpoint import AccountEndpoint
    from pyhuntress.endpoints.siem.ActorEndpoint import ActorEndpoint
    from pyhuntress.endpoints.siem.AgentsEndpoint import AgentsEndpoint
    from pyhuntress.endpoints.siem.BillingreportsEndpoint import BillingreportsEndpoint
    from pyhuntress.endpoints.siem.IncidentreportsEndpoint import IncidentreportsEndpoint
    from pyhuntress.endpoints.siem.OrganizationsEndpoint import OrganizationsEndpoint
    from pyhuntress.endpoints.siem.ReportsEndpoint import ReportsEndpoint
    from pyhuntress.endpoints.siem.SignalsEndpoint import SignalsEndpoint


class HuntressSIEMAPIClient(HuntressClient):
    """
    Huntress SIEM API client. Handles the connection to the Huntress SIEM API
    and the configuration of all the available endpoints.
    """

    def __init__(
        self,
        publickey: str,
        privatekey: str,
    ) -> None:
        """
        Initializes the client with the given credentials.

        Parameters:
            publickey (str): Your Huntress SIEM API public key.
            privatekey (str): Your Huntress SIEM API private key.
        """
        self.publickey: str = publickey
        self.privatekey: str = privatekey
        self.token_expiry_time: datetime = datetime.now(tz=timezone.utc)

        # Grab first access token
        self.base64_auth: str = self._get_auth_key()

    # Initializing endpoints
    @property
    def account(self) -> "AccountEndpoint":
        from pyhuntress.endpoints.siem.AccountEndpoint import AccountEndpoint

        return AccountEndpoint(self)

    @property
    def actor(self) -> "ActorEndpoint":
        from pyhuntress.endpoints.siem.ActorEndpoint import ActorEndpoint

        return ActorEndpoint(self)

    @property
    def agents(self) -> "AgentsEndpoint":
        from pyhuntress.endpoints.siem.AgentsEndpoint import AgentsEndpoint

        return AgentsEndpoint(self)

    @property
    def billing_reports(self) -> "BillingreportsEndpoint":
        from pyhuntress.endpoints.siem.BillingreportsEndpoint import BillingreportsEndpoint

        return BillingreportsEndpoint(self)

    @property
    def incident_reports(self) -> "IncidentreportsEndpoint":
        from pyhuntress.endpoints.siem.IncidentreportsEndpoint import IncidentreportsEndpoint

        return IncidentreportsEndpoint(self)

    @property
    def organizations(self) -> "OrganizationsEndpoint":
        from pyhuntress.endpoints.siem.OrganizationsEndpoint import OrganizationsEndpoint

        return OrganizationsEndpoint(self)

    @property
    def reports(self) -> "ReportsEndpoint":
        from pyhuntress.endpoints.siem.ReportsEndpoint import ReportsEndpoint

        return ReportsEndpoint(self)

    @property
    def signals(self) -> "SignalsEndpoint":
        from pyhuntress.endpoints.siem.SignalsEndpoint import SignalsEndpoint

        return SignalsEndpoint(self)

    def _get_url(self) -> str:
        """
        Generates and returns the URL for the Huntress SIEM API endpoints based on the company url and codebase.
        Logs in an obtains an access token.
        Returns:
            str: API URL.
        """
        return f"https://api.huntress.io/v1"

    def _get_auth_key(self) -> str:
        """
        Creates a base64 encoded authentication string to the Huntress SIEM API to obtain an access token.
        """
        # Format: base64encode(api_key:api_secret)
        
        auth_str = f"{self.publickey}:{self.privatekey}"
        auth_bytes = auth_str.encode('ascii')
        base64_auth = base64.b64encode(auth_bytes).decode('ascii')

        return base64_auth

    def _get_headers(self) -> dict[str, str]:
        """
        Generates and returns the headers required for making API requests. The access token is refreshed if necessary before returning.

        Returns:
            dict[str, str]: Dictionary of headers including Content-Type, Client ID, and Authorization.
        """
        return {
            "Content-Type": "application/json",
            "Authorization": f"Basic {self.base64_auth}",
        }



import re
import math
from datetime import datetime
from typing import Any

from requests.structures import CaseInsensitiveDict


def cw_format_datetime(dt: datetime) -> str:
    """Format a datetime object as a string in ISO 8601 format. This is the format that Huntress uses.

    Args:
        dt (datetime): The datetime object to be formatted.

    Returns:
        str: The formatted datetime string in the format "YYYY-MM-DDTHH:MM:SSZ".

    Example:
        from datetime import datetime

        dt = datetime(2022, 1, 1, 12, 0, 0)
        formatted_dt = cw_format_datetime(dt)
        print(formatted_dt)  # Output: "2022-01-01T12:00:00Z"
    """
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")

def parse_response_body(
    body: CaseInsensitiveDict,
) -> dict[str, Any] | None:
    """
    Parses response body to extract pagination information.

    Arguments:
    - body: content.json().get('pagination', {}) A dictionary containing the headers of an HTTP response.

    Returns:
    - A dictionary containing the extracted pagination information. The keys in the dictionary include:
      - "first_page": An optional integer representing the number of the first page.
      - "prev_page": An optional integer representing the number of the previous page.
      - "next_page": An optional integer representing the number of the next page.
      - "last_page": An optional integer representing the number of the last page.
      - "has_next_page": A boolean indicating whether there is a next page.
      - "has_prev_page": A boolean indicating whether there is a previous page.

    If the "Link" header is not present in the headers dictionary, None is returned.

    Example Usage:
        headers = {
            "Link": '<https://example.com/api?page=1>; rel="first", <https://example.com/api?page=2>; rel="next"'
        }
        pagination_info = parse_link_headers(headers)
        print(pagination_info)
        # Output: {'first_page': 1, 'next_page': 2, 'has_next_page': True}
    """
    if body.get("current_page") is None:
        return None
    has_next_page: bool = False
    has_prev_page: bool = False
    first_page: int | None = None
    prev_page: int | None = None
    current_page: int | None = None
    current_page_count: int | None = None
    limit: int | None = None
    total_count: int | None = None
    next_page: int | None = None
    next_page_url: str | None = None
    next_page_token: str | None = None
    last_page: int | None = None

    result = {}

    if body.get("first_page") is not None:
        result["first_page"] = body.get("first_page")

    if body.get("prev_page") is not None:
        result["prev_page"] = body.get("prev_page")
    elif body.get("current_page") is not None:
        if body.get("current_page") > 1:
            result["prev_page"] = body.get("current_page") - 1
    elif body.get("currentPage") is not None:
        if body.get("currentPage") > 1:
            result["prev_page"] = body.get("currentPage") - 1

    if body.get("next_page") is not None:
        result["next_page"] = body.get("next_page")
    elif body.get("currentPage") is not None and body.get("currentPage") < body.get("lastPage"):
        result["next_page"] = body.get("currentPage") + 1

    if body.get("last_page") is not None:
        result["last_page"] = body.get("last_page")
    elif body.get("lastPage") is not None:
        result["last_page"] = body.get("lastPage")
    elif body.get("last_page") is None and body.get("current_page") is not None:
        result["last_page"] = math.ceil(body.get("total_count")/body.get("limit"))

    if body.get("has_next_page"):
        result["has_next_page"] = body.get("has_next_page")
    elif body.get("current_page") is not None and body.get("next_page") is not None:
        result["has_next_page"] = True
    elif body.get("current_page") is not None and body.get("next_page") is None:
        result["has_next_page"] = False
    elif body.get("currentPage") is not None and body.get("currentPage") < body.get("lastPage"):
        result["has_next_page"] = True
    
    if body.get("has_prev_page"):
        result["has_prev_page"] = body.get("has_prev_page")
    elif body.get("current_page") is not None:
        if body.get("current_page") > 1:
            result["has_prev_page"] = True
    elif body.get("currentPage") is not None:
        if body.get("currentPage") > 1:
            result["has_prev_page"] = True

    return result
    
def parse_link_headers(
    headers: CaseInsensitiveDict,
) -> dict[str, Any] | None:
    """
    Parses link headers to extract pagination information.

    Arguments:
    - headers: A dictionary containing the headers of an HTTP response. The value associated with the "Link" key should be a string representing the link headers.

    Returns:
    - A dictionary containing the extracted pagination information. The keys in the dictionary include:
      - "first_page": An optional integer representing the number of the first page.
      - "prev_page": An optional integer representing the number of the previous page.
      - "next_page": An optional integer representing the number of the next page.
      - "last_page": An optional integer representing the number of the last page.
      - "has_next_page": A boolean indicating whether there is a next page.
      - "has_prev_page": A boolean indicating whether there is a previous page.

    If the "Link" header is not present in the headers dictionary, None is returned.

    Example Usage:
        headers = {
            "Link": '<https://example.com/api?page=1>; rel="first", <https://example.com/api?page=2>; rel="next"'
        }
        pagination_info = parse_link_headers(headers)
        print(pagination_info)
        # Output: {'first_page': 1, 'next_page': 2, 'has_next_page': True}
    """
    if headers.get("Link") is None:
        return None
    links = headers["Link"].split(",")
    has_next_page: bool = False
    has_prev_page: bool = False
    first_page: int | None = None
    prev_page: int | None = None
    next_page: int | None = None
    last_page: int | None = None

    for link in links:
        match = re.search(r'page=(\d+)>; rel="(.*?)"', link)
        if match:
            page_number = int(match.group(1))
            rel_value = match.group(2)
            if rel_value == "first":
                first_page = page_number
            elif rel_value == "prev":
                prev_page = page_number
                has_prev_page = True
            elif rel_value == "next":
                next_page = page_number
                has_next_page = True
            elif rel_value == "last":
                last_page = page_number

    result = {}

    if first_page is not None:
        result["first_page"] = first_page

    if prev_page is not None:
        result["prev_page"] = prev_page

    if next_page is not None:
        result["next_page"] = next_page

    if last_page is not None:
        result["last_page"] = last_page

    if has_next_page:
        result["has_next_page"] = has_next_page

    if has_prev_page:
        result["has_prev_page"] = has_prev_page

    return result



import json
from typing import ClassVar
from urllib.parse import urlsplit, urlunsplit

from requests import JSONDecodeError, Response


class HuntressException(Exception):
    _code_explanation: ClassVar[str] = ""  # Ex: for 404 "Not Found"
    _error_suggestion: ClassVar[str] = ""  # Ex: for 404 "Check the URL you are using is correct"

    def __init__(self, req_response: Response, *, extra_message: str = "") -> None:
        self.response = req_response
        self.extra_message = extra_message
        super().__init__(self.message())

    def _get_sanitized_url(self) -> str:
        """
        Simplify URL down to method, hostname, and path.
        """
        url_components = urlsplit(self.response.url)
        return urlunsplit(
            (
                url_components.scheme,
                url_components.hostname,
                url_components.path,
                "",
                "",
            )
        )

    def details(self) -> str:
        try:
            # If response was json, then format it nicely
            return json.dumps(self.response.json(), indent=4)
        except JSONDecodeError:
            return self.response.text

    def message(self) -> str:
        return (
            f"A HTTP {self.response.status_code} ({self._code_explanation}) error has occurred while requesting"
            f" {self._get_sanitized_url()}.\n{self.response.reason}\n{self._error_suggestion}\n{self.extra_message}"
        ).strip()  # Remove extra whitespace (Ex: if extra_message == "")


class MalformedRequestException(HuntressException):
    _code_explanation = "Bad Request"
    _error_suggestion = (
        "The request could not be understood by the server due to malformed syntax. Please check modify your request"
        " before retrying."
    )


class AuthenticationFailedException(HuntressException):
    _code_explanation = "Unauthorized"
    _error_suggestion = "Please check your credentials are correct before retrying."


class PermissionsFailedException(HuntressException):
    _code_explanation = "Forbidden"
    _error_suggestion = "You may be attempting to access a resource you do not have the appropriate permissions for."


class NotFoundException(HuntressException):
    _code_explanation = "Not Found"
    _error_suggestion = "You may be attempting to access a resource that has been moved or deleted."


class MethodNotAllowedException(HuntressException):
    _code_explanation = "Method Not Allowed"
    _error_suggestion = "This resource does not support the HTTP method you are trying to use."


class ConflictException(HuntressException):
    _code_explanation = "Conflict"
    _error_suggestion = "This resource is possibly in use or conflicts with another record."

class TooManyRequestsException(HuntressException):
    _code_explanation = "Too Many Requests"
    _error_suggestion = "This resource is currently being rate limited. Please wait and try again."


class ServerError(HuntressException):
    _code_explanation = "Internal Server Error"


class ObjectExistsError(HuntressException):
    _code_explanation = "Object Exists"
    _error_suggestion = "This resource already exists."



from pyhuntress.endpoints.base.huntress_endpoint import HuntressEndpoint
from pyhuntress.endpoints.managedsat.PhishingAttemptsActionsEndpoint import PhishingAttemptsActionsEndpoint
from pyhuntress.interfaces import (
    IPostable,
)
from pyhuntress.models.managedsat import SATPhishingAttemptsReport
from pyhuntress.types import (
    JSON,
    HuntressSATRequestParams,
)


class PhishingAttemptsActionsEndpoint(
    HuntressEndpoint,
    IPostable[SATPhishingAttemptsReport, HuntressSATRequestParams],
):
    def __init__(self, client, parent_endpoint=None) -> None:
        HuntressEndpoint.__init__(self, client, "phishing-attempts", parent_endpoint=parent_endpoint)
        IPostable.__init__(self, SATPhishingAttemptsReport)

        self.actions = self._register_child_endpoint(PhishingAttemptsActionsEndpoint(client, parent_endpoint=self))



from pyhuntress.clients.managedsat_client import HuntressSATAPIClient
from pyhuntress.clients.siem_client import HuntressSIEMAPIClient

__all__ = ["HuntressSATAPIClient", "HuntressSIEMAPIClient"]
__version__ = "0.1.1"



from __future__ import annotations

import inspect
from types import UnionType
from typing import Union, get_args, get_origin

from pydantic import BaseModel, ConfigDict

from pyhuntress.utils.naming import to_camel_case


class HuntressModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel_case,
        populate_by_name=True,
        use_enum_values=True,
        protected_namespaces=(),
    )

    @classmethod
    def _get_field_names(cls) -> list[str]:
        field_names = []
        for v in cls.__fields__.values():
            was_model = False
            for arg in get_args(v.annotation):
                if inspect.isclass(arg) and issubclass(arg, HuntressModel):
                    was_model = True
                    field_names.extend([f"{v.alias}/{sub}" for sub in arg._get_field_names()])

            if not was_model:
                field_names.append(v.alias)

        return field_names

    @classmethod
    def _get_field_names_and_types(cls) -> dict[str, str]:  # noqa: C901
        field_names_and_types = {}
        for v in cls.__fields__.values():
            was_model = False
            field_type = "None"
            if get_origin(v.annotation) is UnionType or get_origin(v.annotation) is Union:
                for arg in get_args(v.annotation):
                    if inspect.isclass(arg) and issubclass(arg, HuntressModel):
                        was_model = True
                        for sk, sv in arg._get_field_names_and_types().items():
                            field_names_and_types[f"{v.alias}/{sk}"] = sv
                    elif arg is not None and arg.__name__ != "NoneType":
                        field_type = arg.__name__
            else:
                if inspect.isclass(v.annotation) and issubclass(v.annotation, HuntressModel):
                    was_model = True
                    for sk, sv in v.annotation._get_field_names_and_types().items():
                        field_names_and_types[f"{v.alias}/{sk}"] = sv
                elif v.annotation is not None and v.annotation.__name__ != "NoneType":
                    field_type = v.annotation.__name__

            if not was_model:
                field_names_and_types[v.alias] = field_type

        return field_names_and_types



from __future__ import annotations
import json

from typing import TYPE_CHECKING, Generic, TypeVar

from pyhuntress.utils.helpers import parse_link_headers, parse_response_body

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pydantic import BaseModel
    from requests import Response

    from pyhuntress.types import RequestParams


TModel = TypeVar("TModel", bound="BaseModel")

if TYPE_CHECKING:
    from pyhuntress.interfaces import IPaginateable


class PaginatedResponse(Generic[TModel]):
    """
    PaginatedResponse is a wrapper class for handling paginated responses from the
    Huntress API. It provides methods for navigating through the pages of the response
    and accessing the data contained within each page.

    The class is designed to work with HuntressEndpoint and its derived classes to
    parse the API response into model instances. It also supports iteration, allowing
    the user to loop through the items within the paginated response.

    PaginatedResponse uses a generic type variable TModel, which represents the
    expected model type for the response data. This allows for type-safe handling
    of model instances throughout the class.
    """

    def __init__(
        self,
        response: Response,
        response_model: type[TModel],
        endpointmodel: IPaginateable,
        endpoint: str,
        page: int,
        limit: int,
        params: RequestParams | None = None,
    ) -> None:
        """
        PaginatedResponse is a wrapper class for handling paginated responses from the
        Huntress API. It provides methods for navigating through the pages of the response
        and accessing the data contained within each page.

        The class is designed to work with HuntressEndpoint and its derived classes to
        parse the API response into model instances. It also supports iteration, allowing
        the user to loop through the items within the paginated response.

        PaginatedResponse uses a generic type variable TModel, which represents the
        expected model type for the response data. This allows for type-safe handling
        of model instances throughout the class.
        """
        self._initialize(response, response_model, endpointmodel, endpoint, page, limit, params)

    def _initialize(
        self,
        response: Response,
        response_model: type[TModel],
        endpointmodel: IPaginateable,
        endpoint: str,
        page: int,
        limit: int,
        params: RequestParams | None = None,
    ):
        """
        Initialize the instance variables using the provided response, endpointmodel, and page size.

        Args:
            response: The raw response object from the API.
            endpointmodel (HuntressEndpoint[TModel]): The endpointmodel associated with the response.
            endpoint: The endpoint url to extract the data
            limit (int): The number of items per page.
        """
        self.response = response
        self.response_model = response_model
        self.endpointmodel = endpointmodel
        self.endpoint = endpoint
        self.limit = limit
        # Get page data from the response body
        try:
            self.parsed_pagination_response = parse_response_body(json.loads(response.content.decode('utf-8')).get('pagination', {}))
        except:
            self.parsed_pagination_response = parse_response_body(json.loads(response.content.decode('utf-8')).get('meta.page', {}))
        self.params = params
        if self.parsed_pagination_response is not None:
            # Huntress SIEM API gives us a handy response to parse for Pagination
            self.has_next_page: bool = self.parsed_pagination_response.get("has_next_page", False)
            self.has_prev_page: bool = self.parsed_pagination_response.get("has_prev_page", False)
            self.first_page: int = self.parsed_pagination_response.get("first_page", None)
            self.prev_page: int = self.parsed_pagination_response.get("prev_page", None)
            self.next_page: int = self.parsed_pagination_response.get("next_page", None)
            self.last_page: int = self.parsed_pagination_response.get("last_page", None)
        else:
            # Huntress Managed SAT might, haven't worked on this yet
            self.has_next_page: bool = True
            self.has_prev_page: bool = page > 1
            self.first_page: int = 1
            self.prev_page = page - 1 if page > 1 else 1
            self.next_page = page + 1
            self.last_page = 999999
        self.data: list[TModel] = [response_model.model_validate(d) for d in response.json().get(endpoint, {})]
        self.has_data = self.data and len(self.data) > 0
        self.index = 0

    def get_next_page(self) -> PaginatedResponse[TModel]:
        """
        Fetch the next page of the paginated response.

        Returns:
            PaginatedResponse[TModel]: The updated PaginatedResponse instance
            with the data from the next page or None if there is no next page.
        """
        if not self.has_next_page or not self.next_page:
            self.has_data = False
            return self

        next_response = self.endpointmodel.paginated(self.next_page, self.limit, self.params)
        self._initialize(
            next_response.response,
            next_response.response_model,
            next_response.endpointmodel,
            next_response.endpoint,
            self.next_page,
            next_response.limit,
            self.params,
        )
        return self

    def get_previous_page(self) -> PaginatedResponse[TModel]:
        """
        Fetch the next page of the paginated response.

        Returns:
            PaginatedResponse[TModel]: The updated PaginatedResponse instance
            with the data from the next page or None if there is no next page.
        """
        if not self.has_prev_page or not self.prev_page:
            self.has_data = False
            return self

        prev_response = self.endpointmodel.paginated(self.prev_page, self.limit, self.params)
        self._initialize(
            prev_response.response,
            prev_response.response_model,
            prev_response.endpointmodel,
            self.prev_page,
            prev_response.limit,
            self.params,
        )
        return self

    def all(self) -> Iterable[TModel]:
        """
        Iterate through all items in the paginated response, across all pages.

        Yields:
            TModel: An instance of the model class for each item in the paginated response.
        """
        while self.has_data:
            yield from self.data
            self.get_next_page()

    def __iter__(self):
        """
        Implement the iterator protocol for the PaginatedResponse class.

        Returns:
            PaginatedResponse[TModel]: The current instance of the PaginatedResponse.
        """
        return self

    def __dict__(self):
        """
        Implement the iterator protocol for the PaginatedResponse class.

        Returns:
            PaginatedResponse[TModel]: The current instance of the PaginatedResponse.
        """
        return self.data

    def __next__(self):
        """
        Implement the iterator protocol by getting the next item in the data.

        Returns:
            TModel: The next item in the data.

        Raises:
            StopIteration: If there are no more items in the data.
        """
        if self.index < len(self.data):
            result = self.data[self.index]
            self.index += 1
            return result
        else:
            raise StopIteration



from __future__ import annotations

import contextlib
import json
import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, cast

import requests
from requests import Response
from requests.exceptions import Timeout

from pyhuntress.config import Config
from pyhuntress.exceptions import (
    AuthenticationFailedException,
    ConflictException,
    MalformedRequestException,
    MethodNotAllowedException,
    NotFoundException,
    ObjectExistsError,
    PermissionsFailedException,
    ServerError,
    TooManyRequestsException,
)

if TYPE_CHECKING:
    from pyhuntress.types import RequestData, RequestMethod, RequestParams


class HuntressClient(ABC):
    config: Config = Config()

    @abstractmethod
    def _get_headers(self) -> dict[str, str]:
        pass

    @abstractmethod
    def _get_url(self) -> str:
        pass

    def _make_request(  # noqa: C901
        self,
        method: RequestMethod,
        url: str,
        data: RequestData | None = None,
#        rawdata: RequestData | None = None,
        params: RequestParams | None = None,
        headers: dict[str, str] | None = None,
        retry_count: int = 0,
        stream: bool = False,  # noqa: FBT001, FBT002
    ) -> Response:
        """
        Make an API request using the specified method, endpoint, data, and parameters.
        This function isn't intended for use outside of this class.
        Please use the available CRUD methods as intended.

        Args:
            method (str): The HTTP method to use for the request (e.g., GET, POST, PUT, etc.).
            endpoint (str, optional): The endpoint to make the request to.
            data (dict, optional): The request data to send.
            params (dict, optional): The query parameters to include in the request.

        Returns:
            The Response object (see requests.Response).

        Raises:
            Exception: If the request returns a status code >= 400.
        """

        if not headers:
            headers = self._get_headers()

        # I don't like having to cast the params to a dict, but it's the only way I can get mypy to stop complaining about the type.
        # TypedDicts aren't compatible with the dict type and this is the best way I can think of to handle this.
        if data:
            response = requests.request(
                method,
                url,
                headers=headers,
                data=data,
                params=cast(dict[str, Any], params or {}),
                stream=stream,
            )
        else:
            response = requests.request(
                method,
                url,
                headers=headers,
                params=cast(dict[str, Any], params or {}),
                stream=stream,
            )
        if not response.ok:
            with contextlib.suppress(json.JSONDecodeError):
                details: dict = response.json()
                if response.status_code == 400:
                    if details.get("code") == "InvalidObject":
                        errors = details.get("errors", [])
                        if len(errors) > 1:
                            warnings.warn(
                                "Found multiple errors - we may be masking some important error details.  Please submit a Github issue with response.status_code and response.content so we can improve this error handling.",
                                stacklevel=1,
                            )
                        for error in errors:
                            if error.get("code") == "ObjectExists":
                                error.pop("code")  # Don't need code in message
                                raise ObjectExistsError(response, extra_message=json.dumps(error, indent=4))

            if response.status_code == 400:
                raise MalformedRequestException(response)
            if response.status_code == 401:
                raise AuthenticationFailedException(response)
            if response.status_code == 403:
                raise PermissionsFailedException(response)
            if response.status_code == 404:
                raise NotFoundException(response)
            if response.status_code == 405:
                raise MethodNotAllowedException(response)
            if response.status_code == 409:
                raise ConflictException(response)
            if response.status_code == 429:
                raise TooManyRequestsException(response)
            if response.status_code == 500:
                # if timeout is mentioned anywhere in the response then we'll retry.
                # Ideally we'd return immediately on any non-timeout errors (since
                # retries won't help much there), but err towards classifying too much
                # as retries instead of too little.
                if "timeout" in (response.text + response.reason).lower():
                    if retry_count < self.config.max_retries:
                        retry_count += 1
                        return self._make_request(method, url, data, params, headers, retry_count)
                    raise Timeout(response=response)
                raise ServerError(response)

        return response



from keyword import iskeyword


def to_snake_case(string: str) -> str:
    return ("_" if string.startswith("_") else "") + "".join(
        ["_" + i.lower() if i.isupper() else i for i in string.lstrip("_")]
    ).lstrip("_")


def to_camel_case(string: str) -> str:
    string_split = string.split("_")
    return string_split[0] + "".join(word.capitalize() for word in string_split[1:])


def to_title_case_preserve_case(string: str) -> str:
    return string[:1].upper() + string[1:]


def ensure_not_reserved(string: str) -> str:
    if iskeyword(string):
        return string + "_"
    else:  # noqa: RET505
        return string



from pyhuntress.endpoints.base.huntress_endpoint import HuntressEndpoint
from pyhuntress.endpoints.siem.SignalsIdEndpoint import SignalsIdEndpoint
from pyhuntress.interfaces import (
    IGettable,
    IPaginateable,
)
from pyhuntress.models.siem import SIEMSignals
from pyhuntress.responses.paginated_response import PaginatedResponse
from pyhuntress.types import (
    JSON,
    HuntressSIEMRequestParams,
)


class SignalsEndpoint(
    HuntressEndpoint,
    IGettable[SIEMSignals, HuntressSIEMRequestParams],
    IPaginateable[SIEMSignals, HuntressSIEMRequestParams],
):
    def __init__(self, client, parent_endpoint=None) -> None:
        HuntressEndpoint.__init__(self, client, "signals", parent_endpoint=parent_endpoint)
        IGettable.__init__(self, SIEMSignals)
        IPaginateable.__init__(self, SIEMSignals)

    def id(self, id: int) -> SignalsIdEndpoint:
        """
        Sets the ID for this endpoint and returns an initialized SignalsIdEndpoint object to move down the chain.

        Parameters:
            id (int): The ID to set.
        Returns:
            SignalsIdEndpoint: The initialized SignalsIdEndpoint object.
        """
        child = SignalsIdEndpoint(self.client, parent_endpoint=self)
        child._id = id
        return child

    def paginated(
        self,
        page: int,
        limit: int,
        params: HuntressSIEMRequestParams | None = None,
    ) -> PaginatedResponse[SIEMSignals]:
        """
        Performs a GET request against the /signals endpoint and returns an initialized PaginatedResponse object.

        Parameters:
            page (int): The page number to request.
            limit (int): The number of results to return per page.
            params (dict[str, int | str]): The parameters to send in the request query string.
        Returns:
            PaginatedResponse[SIEMSignals]: The initialized PaginatedResponse object.
        """
        if params:
            params["page"] = page
            params["limit"] = limit
        else:
            params = {"page": page, "limit": limit}
        return PaginatedResponse(
            super()._make_request("GET", params=params),
            SIEMSignals,
            self,
            "signals",
            page,
            limit,
            params,
        )

    def get(
        self,
        data: JSON | None = None,
        params: HuntressSIEMRequestParams | None = None,
    ) -> SIEMSignals:
        """
        Performs a GET request against the /signals endpoint.

        Parameters:
            data (dict[str, Any]): The data to send in the request body.
            params (dict[str, int | str]): The parameters to send in the request query string.
        Returns:
            SIEMAuthInformation: The parsed response data.
        """
        return self._parse_many(
            SIEMSignals,
            super()._make_request("GET", data=data, params=params).json().get('signals', {}),
        )
